# -*- coding: utf-8 -*-
#
# Copyright 2014-2015 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""Options for BigMLer analyze

"""
from __future__ import absolute_import


import os


ACCURACY = "accuracy"
BIGML_MAX_NODES = os.environ.get("BIGML_MAX_NODES", 2000)
MINIMIZE_OPTIONS = ["mean_squared_error", "mean_absolute_error"]
MAXIMIZE_OPTIONS = [ACCURACY, "precision", "recall", "phi", "f_measure",
                    "r_squared", "phi_coefficient"]
OPTIMIZE_OPTIONS = [ACCURACY, "precision", "recall", "phi", "f_measure",
                    "mean_squared_error", "mean_absolute_error", "r_squared",
                    "phi_coefficient"]

def get_analyze_options(defaults=None):
    """Adding arguments for the analyze subcommand

    """


    if defaults is None:
        defaults = {}

    options = {
        # K-fold cross-validation flag
        '--cross-validation': {
            "action": 'store_true',
            "dest": 'cv',
            "default": defaults.get('cv', False),
            "help": "K-fold cross-validation flag."},

        # K-fold cross-validation index
        '--k-folds': {
            "action": 'store',
            "dest": 'k_folds',
            "default": defaults.get('k_folds', None),
            "type": int,
            "help": "K-fold cross-validation index."},

        # Name of the directory where session files will be stored. If --output
        # is set, this setting will be overriden by it.
        '--output-dir': {
            "action": 'store',
            "dest": 'output_dir',
            "default": defaults.get('output_dir', None),
            "help": ("Directory where session files will be"
                     " stored. --output file path will "
                     " override it if both"
                     " are set.")},

        # Staleness to stop --features analysis
        '--staleness': {
            "action": 'store',
            "dest": 'staleness',
            "type": float,
            "default": defaults.get('staleness', None),
            "help": ("Limit staleness when using --features"
                     " analysis.")},

        # Penalty per feature in --features analysis
        '--penalty': {
            "action": 'store',
            "dest": 'penalty',
            "type": float,
            "default": defaults.get('penalty', None),
            "help": ("Penalty for each feature in --features"
                     " analysis.")},

        # Features: Set on the feature exclusion analysis
        '--features': {
            "action": 'store_true',
            "dest": 'features',
            "default": defaults.get('features', False),
            "help": "Features analysis."},

        # Mazimize: evaluation measure to be maximized
        '--maximize': {
            "action": 'store',
            "dest": 'maximize',
            "choices": MAXIMIZE_OPTIONS,
            "default": defaults.get('maximize', ACCURACY),
            "help": ("Evaluation measure to be maximized. This flag will be"
                     " deprecated. Please, use --optimize.")},

        # Optimize: evaluation measure to be optimized
        '--optimize': {
            "action": 'store',
            "dest": 'optimize',
            "choices": OPTIMIZE_OPTIONS,
            "default": defaults.get('optimize', ACCURACY),
            "help": "Evaluation measure to be optimized."},

        # Optimize-category: in categorical models, name of the category
        # whose evaluation measure is to be optimized
        '--optimize-category': {
            "action": 'store',
            "dest": 'optimize_category',
            "default": defaults.get('optimize-category', None),
            "help": ("Name of the category whose evaluation measure"
                     " is to be optimized.")},

        # Min_nodes: Minimum number of nodes to start the node threshold
        # analysis
        '--min-nodes': {
            "action": 'store',
            "dest": 'min_nodes',
            "type": int,
            "default": defaults.get('min_nodes', None),
            "help": ("Minimum number of nodes to start the"
                     " node threshold analysis.")},

        # Max_nodes: Maximum number of nodes to end the node threshold analysis
        '--max-nodes': {
            "action": 'store',
            "dest": 'max_nodes',
            "type": int,
            "default": defaults.get('max_nodes',
                                    BIGML_MAX_NODES),
            "help": ("Maximum number of nodes to end the"
                     " node threshold analysis.")},

        # Nodes: Set on node threshold analysis.
        '--nodes': {
            "action": 'store_true',
            "dest": 'nodes',
            "default": defaults.get('nodes', False),
            "help": "Node threshold analysis."},

        # Nodes step: Step to increase the nodes threshold number
        '--nodes-step': {
            "action": 'store',
            "dest": 'nodes_step',
            "type": int,
            "default": defaults.get('nodes_step', None),
            "help": ("Step to increase the nodes"
                     " threshold number in nodes "
                     " threshold analysis. If not set,"
                     " an increase of 100 is used")},

        # Exclude some features from the features analyze
        '--exclude-features': {
            "action": 'store',
            "dest": 'exclude_features',
            "default": defaults.get('exclude_features',
                                    None),
            "help": ("List of comma-separated field names to"
                     " be excluded from the features"
                     " analysis.")},

        # random candidates: Set on random candidates analysis.
        '--random-fields': {
            "action": 'store_true',
            "dest": 'random_fields',
            "default": defaults.get('random_fields', False),
            "help": "Random candidates analysis for ensembles."}}

    return options



# -*- coding: utf-8 -*-
#
# Copyright 2014-2015 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Common options for the BigMLer parser

"""
from __future__ import absolute_import

import datetime

def get_common_options(defaults=None, constants=None):
    """Common options used in all subcommands

    """

    if defaults is None:
        defaults = {}
    if constants is None:
        constants = {}
    now = constants.get('NOW',
                        datetime.datetime.now().strftime("%a%b%d%y_%H%M%S"))
    version_text = constants['version_text']
    defaults_tag = defaults.get('tag')
    defaults_tag = [] if defaults_tag is None else defaults_tag.split(",")
    options = {
        # Shows log info for each https request.
        '--debug': {
            "action": 'store_true',
            "default": defaults.get('debug', False),
            "help": "Activate debug level"},

        # Uses BigML dev environment. Sizes must be under 16MB though.
        '--dev': {
            "action": 'store_true',
            "dest": 'dev_mode',
            "default": defaults.get('dev', False),
            "help": ("Compute a test output using BigML FREE"
                     " development environment.")},

        # BigML's username.
        '--username': {
            "action": 'store',
            "default": defaults.get('username', None),
            "help": "BigML's username."},

        # BigML's API key.
        '--api-key': {
            "action": 'store',
            "dest": 'api_key',
            "default": defaults.get('api_key', None),
            "help": "BigML's API key."},

        # Log file to store resources ids.
        '--resources-log': {
            "action": 'store',
            "dest": 'log_file',
            "default": defaults.get('resources_log', None),
            "help": ("Path to a file to store new resources"
                     " ids. One resource per line"
                     " (e.g., model/50a206a8035d0706dc000376"
                     ").")},

        # Stores the retrieved resources in the output directory
        '--store': {
            "action": 'store_true',
            "dest": 'store',
            "default": defaults.get('store', False),
            "help": ("Store the retrieved resources in the"
                     " output directory.")},

        # Clear global bigmler log files
        '--clear-logs': {
            "action": 'store_true',
            "dest": 'clear_logs',
            "default": defaults.get('clear_logs', False),
            "help": "Clear global bigmler log files."},

        # Use it to add a tag to the new resources created.
        '--tag': {
            "action": 'append',
            "default": defaults_tag,
            "help": "Tag to later retrieve new resources."},

        # Avoid default tagging of resources.
        '--no-tag': {
            "action": 'store_false',
            "dest": 'no_tag',
            "default": defaults.get('no_tag', True),
            "help": "No tag resources with default BigMLer tags."},

        # Hides log info for each https request.
        '--no-debug': {
            "action": 'store_false',
            "dest": 'debug',
            "default": defaults.get('debug', False),
            "help": "Deactivate debug level."},

        # Uses BigML standard environment.
        '--no-dev': {
            "action": 'store_false',
            "dest": 'dev_mode',
            "default": defaults.get('dev', False),
            "help": ("Compute a test output using BigML "
                     "standard development environment.")},

        # Turn on/off verbosity
        '--verbosity': {
            "action": 'store',
            "dest": 'verbosity',
            "default": defaults.get('verbosity', 1),
            "type": int,
            "choices": [0, 1],
            "help": "Set verbosity: 0 to turn off, 1 to turn on."},

        # Resume a partial execution
        '--resume': {
            "action": 'store_true',
            "help": "Resume command."},

        # Resume a partial execution
        '--stack-level': {
            "action": 'store',
            "dest": 'stack_level',
            "default": 0,
            "type": int,
            "help": "Resume command."},

        # Separator for multiple arguments
        '--args-separator': {
            "action": 'store',
            "dest": 'args_separator',
            "default": defaults.get('args_separator', ","),
            "help": ("Separator used in arguments with "
                     " multiple values.")},

        # Name to be used with the source and then with datasets, models and
        # predictions.
        '--name': {
            "action": 'store',
            "dest": 'name',
            "default": defaults.get('name', 'BigMLer_%s' % now),
            "help": "Name for the resources in BigML."},

        # Name of the directory where session files will be stored. If --output
        # is set, this setting will be overriden by it.
        '--output-dir': {
            'action': 'store',
            'dest': 'output_dir',
            'default': defaults.get('output_dir', None),
            'help': ("Directory where session files will be stored."
                     " --output file path will override it if both"
                     " are set.")},

        # Shared. Shares all shareable resources and uses its shared links in
        # reports
        '--shared': {
            'action': 'store_true',
            'dest': 'shared',
            'default': defaults.get('shared', False),
            'help': ("Share resources and use its shared urls "
                     " in reports.")},

        # Category code.
        '--category': {
            'action': 'store',
            'dest': 'category',
            'default': defaults.get('category', 12),
            'type': int,
            'help': "Category code."},

        # A file including a markdown description
        '--description': {
            'action': 'store',
            'dest': 'description',
            'default': defaults.get('description', None),
            'help': ("Path to a file with a description in plain"
                     " text or markdown.")},

        # Set default tagging of resources. (opposed to --no-tag)
        '--no-no-tag': {
            'action': 'store_true',
            'dest': 'no_tag',
            'default': defaults.get('no_tag', True),
            'help': "Tag resources with default BigMLer tags."},

        # Evaluations flag: excluding one dataset from the datasets list to test
        '--version': {
            'action': 'version',
            'version': version_text}}

    return options



# -*- coding: utf-8 -*-
#
# Copyright 2015 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""BigMLer - sample subcommand processing dispatching

"""
from __future__ import absolute_import

import sys
import os
import shutil

import bigmler.utils as u
import bigmler.resources as r
import bigmler.pre_model_steps as pms
import bigmler.processing.args as a
import bigmler.processing.samples as psa

from bigmler.defaults import DEFAULTS_FILE
from bigmler.command import get_stored_command
from bigmler.dispatcher import (SESSIONS_LOG, command_handling,
                                clear_log_files)
from bigmler.sampleoutput import sample_file
from bigmler.reports import clear_reports, upload_reports

COMMAND_LOG = u".bigmler_sample"
DIRS_LOG = u".bigmler_sample_dir_stack"
LOG_FILES = [COMMAND_LOG, DIRS_LOG, u.NEW_DIRS_LOG]
MINIMUM_MODEL = "full=false"
DEFAULT_OUTPUT = u"sample.csv"


def has_sample(args):
    """Returns if some kind of sample id is given in args.

    """
    return args.sample or args.samples or args.sample_tag


def sample_dispatcher(args=sys.argv[1:]):
    """Parses command line and calls the different processing functions

    """

    # If --clear-logs the log files are cleared
    if "--clear-logs" in args:
        clear_log_files(LOG_FILES)

    command = command_handling(args, COMMAND_LOG)

    # Parses command line arguments.
    command_args = a.parse_and_check(command)
    resume = command_args.resume
    if command_args.resume:
        command_args, session_file, output_dir = get_stored_command(
            args, command_args.debug, command_log=COMMAND_LOG,
            dirs_log=DIRS_LOG, sessions_log=SESSIONS_LOG)
        command_args.predictions = os.path.join(output_dir,
                                                DEFAULT_OUTPUT)
    else:
        if command_args.output_dir is None:
            command_args.output_dir = a.NOW
        command_args.predictions = os.path.join(command_args.output_dir,
                                                DEFAULT_OUTPUT)
        if len(os.path.dirname(command_args.predictions).strip()) == 0:
            command_args.predictions = os.path.join(command_args.output_dir,
                                                    command_args.predictions)
        directory = u.check_dir(command_args.predictions)
        session_file = os.path.join(directory, SESSIONS_LOG)
        u.log_message(command.command + "\n", log_file=session_file)
        try:
            shutil.copy(DEFAULTS_FILE, os.path.join(directory, DEFAULTS_FILE))
        except IOError:
            pass
        u.sys_log_message(u"%s\n" % os.path.abspath(directory),
                          log_file=DIRS_LOG)

    # Creates the corresponding api instance
    api = a.get_api_instance(command_args, u.check_dir(session_file))

    # Selects the action to perform
    if a.has_train(command_args) or has_sample(command_args):
        output_args = a.get_output_args(api, command_args, resume)
        a.transform_args(command_args, command.flags, api,
                         command.user_defaults)
        compute_output(**output_args)
    u.log_message("_" * 80 + "\n", log_file=session_file)


def compute_output(api, args):
    """ Creates a sample based on a `train_set`, source or dataset.

    """

    samples = None
    # variables from command-line options
    resume = args.resume_
    sample_ids = args.sample_ids_
    output = args.predictions
    # there's only one sample to be generated at present
    args.max_parallel_clusters = 1
    # sample cannot be published yet.
    args.public_sample = False

    # It is compulsory to have a description to publish either datasets or
    # clusters
    if (not args.description_ and (args.public_sample or
                                   args.public_dataset)):
        sys.exit("You should provide a description to publish.")

    # When using --new-fields, it is compulsory to specify also a dataset
    # id
    if args.new_fields and not args.dataset:
        sys.exit("To use --new-fields you must also provide a dataset id"
                 " to generate the new dataset from it.")

    path = u.check_dir(output)
    session_file = u"%s%s%s" % (path, os.sep, SESSIONS_LOG)
    csv_properties = {}
    # If logging is required set the file for logging
    log = None
    if args.log_file:
        u.check_dir(args.log_file)
        log = args.log_file
        # If --clear_logs the log files are cleared
        clear_log_files([log])

    # basic pre-sample step: creating or retrieving the source related info
    source, resume, csv_properties, fields = pms.get_source_info(
        api, args, resume, csv_properties, session_file, path, log)
    # basic pre-sample step: creating or retrieving the dataset related info
    dataset_properties = pms.get_dataset_info(
        api, args, resume, source,
        csv_properties, fields, session_file, path, log)
    (_, datasets, _, resume,
     csv_properties, fields) = dataset_properties
    if args.sample_file:
        # sample is retrieved from the contents of the given local JSON file
        sample, csv_properties, fields = u.read_local_resource(
            args.sample_file,
            csv_properties=csv_properties)
        samples = [sample]
        sample_ids = [sample['resource']]
    else:
        # sample is retrieved from the remote object
        samples, sample_ids, resume = psa.samples_processing(
            datasets, samples, sample_ids, api, args, resume,
            session_file=session_file, path=path, log=log)
        if samples:
            sample = samples[0]

    # We update the sample's public state if needed
    if sample:
        if isinstance(sample, basestring):
            # build the query string from the sample options
            sample = u.check_resource(sample, api.get_sample)
        samples[0] = sample
        if (args.public_sample or
                (args.shared_flag and r.shared_changed(args.shared, sample))):
            sample_args = {}
            if args.shared_flag and r.shared_changed(args.shared, sample):
                sample_args.update(shared=args.shared)
            if args.public_sample:
                sample_args.update(r.set_publish_sample_args(args))
            if sample_args:
                sample = r.update_sample(sample, sample_args, args,
                                         api=api, path=path,
                                         session_file=session_file)
                samples[0] = sample

    # We get the fields of the sample if we haven't got
    # them yet and need them
    if sample and psa.needs_sample_fields(args):
        fields = psa.get_sample_fields(sample, csv_properties, args)

    sample_file(samples[0], fields, args, api, path=path,
                session_file=session_file)

    u.print_generated_files(path, log_file=session_file,
                            verbosity=args.verbosity)
    if args.reports:
        clear_reports(path)
        if args.upload:
            upload_reports(args.reports, path)



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Copyright 2015 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


""" Testing batch centroids

"""
from __future__ import absolute_import


from bigmler.tests.world import (world, common_setup_module,
                                 common_teardown_module, teardown_class)


import bigmler.tests.basic_tst_prediction_steps as test_pred
import bigmler.tests.basic_batch_tst_prediction_steps as batch_pred
import bigmler.tests.basic_cluster_prediction_steps as test_cluster
import bigmler.tests.basic_anomaly_prediction_steps as test_anomaly


def setup_module():
    """Setup for the module

    """
    common_setup_module()

def teardown_module():
    """Teardown for the module

    """
    common_teardown_module()

class TestBatchCentroids(object):

    def teardown(self):
        """Calling generic teardown for every method

        """
        print "\nEnd of tests in: %s\n-------------------\n" % __name__
        teardown_class()

    def setup(self):
        """
            Debug information
        """
        print "\n-------------------\nTests in: %s\n" % __name__

    def test_scenario1(self):
        """
            Scenario: Successfully building test centroid predictions from scratch:
                Given I create BigML resources uploading train "<data>" file to find centroids for "<test>" remotely with mapping file "<fields_map>" and log predictions in "<output>"
                And I check that the source has been created
                And I check that the dataset has been created
                And I check that the cluster has been created
                And I check that the source has been created from the test file
                And I check that the dataset has been created from the test file
                And I check that the batch centroid prediction has been created
                And I check that the centroids are ready
                Then the local centroids file is like "<predictions_file>"

                Examples:
                | data               | test                    | fields_map | output                        |predictions_file           |
                | ../data/grades.csv | ../data/grades_perm.csv | ../data/grades_fields_map_perm.csv | ./scenario_cb_1_r/centroids.csv | ./check_files/centroids_grades.csv |
        """
        print self.test_scenario1.__doc__
        examples = [
            ['data/grades.csv', 'data/grades_perm.csv', 'data/grades_fields_map_perm.csv', 'scenario_cb_1_r/centroids.csv', 'check_files/centroids_grades.csv']]
        for example in examples:
            print "\nTesting with:\n", example
            test_cluster.i_create_all_cluster_resources_with_mapping(self, data=example[0], test=example[1], fields_map=example[2], output=example[3])
            test_pred.i_check_create_source(self)
            test_pred.i_check_create_dataset(self, suffix=None)
            test_pred.i_check_create_cluster(self)
            test_pred.i_check_create_test_source(self)
            test_pred.i_check_create_test_dataset(self)
            batch_pred.i_check_create_batch_centroid(self)
            test_cluster.i_check_create_centroids(self)
            test_pred.i_check_predictions(self, example[4])

    def test_scenario2(self):
        """
            Scenario: Successfully building remote test centroid predictions from scratch to dataset:
                Given I create BigML resources uploading train "<data>" file to find centroids for "<test>" remotely to dataset with no CSV and log resources in "<output_dir>"
                And I check that the source has been created
                And I check that the dataset has been created
                And I check that the cluster has been created
                And I check that the source has been created from the test file
                And I check that the dataset has been created from the test file
                And I check that the batch centroid prediction has been created
                Then I check that the batch centroids dataset exists
                And no local CSV file is created

                Examples:
                | data               | test                    |  output_dir     |
                | ../data/grades.csv | ../data/test_grades.csv | ./scenario_cb_2 |

        """
        print self.test_scenario2.__doc__
        examples = [
            ['data/grades.csv', 'data/test_grades.csv', 'scenario_cb_2']]
        for example in examples:
            print "\nTesting with:\n", example
            test_cluster.i_create_all_cluster_resources_to_dataset(self, data=example[0], test=example[1], output_dir=example[2])
            test_pred.i_check_create_source(self)
            test_pred.i_check_create_dataset(self, suffix=None)
            test_pred.i_check_create_cluster(self)
            test_pred.i_check_create_test_source(self)
            test_pred.i_check_create_test_dataset(self)
            batch_pred.i_check_create_batch_centroid(self)
            batch_pred.i_check_create_batch_centroids_dataset(self)
            test_anomaly.i_check_no_local_CSV(self)


    def test_scenario3(self):
        """
            Scenario: Successfully building remote test centroid predictions from scratch with prediction fields:
                Given I create BigML resources uploading train "<data>" file to find centroids for "<test>" remotely with prediction fields "<prediction_fields>" and log resources in "<output>"
                And I check that the source has been created
                And I check that the dataset has been created
                And I check that the cluster has been created
                And I check that the source has been created from the test file
                And I check that the dataset has been created from the test file
                And I check that the batch centroid prediction has been created
                And I check that the centroids are ready
                Then the local centroids file is like "<predictions_file>"

                Examples:
                | data               | test                    |  prediction_fields | output     | predictions_file
                | ../data/grades.csv | ../data/test_grades.csv |  Assignment       |./scenario_cb_3_r/centroids.csv | ./check_files/centroids_grades_field.csv |

        """
        print self.test_scenario3.__doc__
        examples = [
            ['data/grades.csv', 'data/grades.csv', 'Assignment',
             './scenario_cb_3_r/centroids.csv', "./check_files/centroids_grades_field.csv"]]
        for example in examples:
            print "\nTesting with:\n", example
            test_cluster.i_create_all_cluster_resources_with_prediction_fields(
                self, data=example[0], test=example[1],
                prediction_fields=example[2], output=example[3])
            test_pred.i_check_create_source(self)
            test_pred.i_check_create_dataset(self, suffix=None)
            test_pred.i_check_create_cluster(self)
            test_pred.i_check_create_test_source(self)
            test_pred.i_check_create_test_dataset(self)
            batch_pred.i_check_create_batch_centroid(self)
            test_cluster.i_check_create_centroids(self)
            test_pred.i_check_predictions(self, example[4])



# -*- coding: utf-8 -*-

from __future__ import absolute_import

__version__ = '3.5.4'

if __name__ == '__main__':
    import sys
    from bigmler.bigmler import main
    sys.exit(main(sys.argv[1:]))



# -*- coding: utf-8 -*-
#
# Copyright 2014-2015 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import


import os
import time
import json
from bigmler.tests.world import world, res_filename
from subprocess import check_call, CalledProcessError
from bigmler.checkpoint import file_number_of_lines
from bigmler.utils import SYSTEM_ENCODING, PYTHON3, open_mode
from bigml.api import check_resource
from bigmler.tests.common_steps import check_debug


#@step(r'I create a new sample from the dataset and get the sample using options "(.*)" storing logs in "(.*)"')
def i_create_sample(step, options=None, output_dir=None):
    if options is None or output_dir is None:
        assert False
    world.directory = output_dir
    world.folders.append(world.directory)
    try:
        command = (u"bigmler sample --dataset " + world.dataset['resource'] +
                   u" --store --output-dir " + output_dir +
                   u" " + options)
        command = check_debug(command)
        if not PYTHON3:
            command.encode(SYSTEM_ENCODING)
        retcode = check_call(command, shell=True)
        if retcode < 0:
            assert False
        else:
            world.output = output_dir
            assert True
    except (OSError, CalledProcessError, IOError) as exc:
        assert False, str(exc)


#@step(r'I check that the sample has been created')
def i_check_create_sample(step):
    sample_file = "%s%ssamples" % (world.directory, os.sep)
    try:
        sample_file = open(sample_file, "r")
        sample = check_resource(sample_file.readline().strip(),
                                world.api.get_sample)
        world.samples.append(sample['resource'])
        world.sample = sample
        sample_file.close()
        assert True
    except Exception, exc:
        assert False, str(exc)


#@step(r'the sample contains stat-info like in "(.*)"')
def i_check_sample_stat(step, check_sample_json=None):
    if check_sample_json is None:
        assert False
    check_sample_json = res_filename(check_sample_json)
    try:
        sample_json = os.path.join(world.directory, "stat_info.json")
        with open(check_sample_json, "r") as check_sample_file:
            check_sample_contents = check_sample_file.read()
        with open(sample_json, "r") as sample_file:
            sample_file_contents = sample_file.read()
        if check_sample_contents == sample_file_contents:
            assert True
        else:
            assert False, ("File contents:\n%s\nExpected contents:\n%s" %
                           (sample_file_contents, check_sample_contents))
    except Exception, exc:
        assert False, str(exc)


#@step(r'the sample file is like "(.*)"')
def i_check_sample_file(step, check_sample_file=None):
    if check_sample_file is None:
        assert False
    check_sample_file = res_filename(check_sample_file)
    try:
        sample_file = os.path.join(world.directory, "sample.csv")
        with open(check_sample_file, open_mode("r")) as check_sample_file:
            check_sample_contents = check_sample_file.read()
        with open(sample_file, open_mode("r")) as sample_file:
            sample_file_contents = sample_file.read()
        if check_sample_contents == sample_file_contents:
            assert True
        else:
            assert False, ("File contents:\n%s\nExpected contents:\n%s" %
                           (sample_file_contents, check_sample_contents))
    except Exception, exc:
        assert False, str(exc)


#@step(r'the sample JSON is like the one in "(.*)"')
def i_check_sample_json(step, check_sample_file=None):
    if check_sample_file is None:
        assert False
    check_sample_file = res_filename(check_sample_file)
    try:
        sample_file = os.path.join(world.directory, "stat_info.json")
        with open(check_sample_file, open_mode("r")) as check_sample_file:
            contents = check_sample_file.read()
            check_sample_json = json.loads(contents)
        with open(sample_file, open_mode("r")) as sample_file:
            contents = sample_file.read()
            sample_file_json = json.loads(contents)
        if check_sample_json == sample_file_json:
            assert True
        else:
            assert False, ("File contents:\n%s\nExpected contents:\n%s" %
                           (sample_file_json, check_sample_json))
    except Exception, exc:
        assert False, str(exc)



# -*- coding: utf-8 -*-
#
# Copyright 2012-2015 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""Defaults parser for BigMLer

"""
import ConfigParser

DEFAULTS_FILE = 'bigmler.ini'

FLAGS = {
    'BigMLer': [
        {'flag': 'debug', 'type': 'boolean'},
        {'flag': 'dev', 'type': 'boolean'},
        {'flag': 'username', 'type': 'string'},
        {'flag': 'api_key', 'type': 'string'},
        {'flag': 'train', 'type': 'string'},
        {'flag': 'test', 'type': 'string'},
        {'flag': 'output', 'type': 'string'},
        {'flag': 'objective', 'type': 'string'},
        {'flag': 'category', 'type': 'int'},
        {'flag': 'description', 'type': 'string'},
        {'flag': 'field_names', 'type': 'string'},
        {'flag': 'field_attributes', 'type': 'string'},
        {'flag': 'types', 'type': 'string'},
        {'flag': 'dataset_fields', 'type': 'string'},
        {'flag': 'json_filter', 'type': 'string'},
        {'flag': 'lisp_filter', 'type': 'string'},
        {'flag': 'model_fields', 'type': 'string'},
        {'flag': 'train_header', 'type': 'boolean'},
        {'flag': 'test_header', 'type': 'boolean'},
        {'flag': 'name', 'type': 'string'},
        {'flag': 'source', 'type': 'string'},
        {'flag': 'dataset', 'type': 'string'},
        {'flag': 'model', 'type': 'string'},
        {'flag': 'model_file', 'type': 'string'},
        {'flag': 'remote', 'type': 'boolean'},
        {'flag': 'models', 'type': 'string'},
        {'flag': 'datasets', 'type': 'string'},
        {'flag': 'pruning', 'type': 'string'},
        {'flag': 'datasets', 'type': 'string'},
        {'flag': 'number_of_models', 'type': 'int'},
        {'flag': 'sample_rate', 'type': 'float'},
        {'flag': 'replacement', 'type': 'boolean'},
        {'flag': 'max_parallel_models', 'type': 'int'},
        {'flag': 'max_batch_models', 'type': 'int'},
        {'flag': 'randomize', 'type': 'boolean'},
        {'flag': 'no_tag', 'type': 'boolean'},
        {'flag': 'tag', 'type': 'string'},
        {'flag': 'model_tag', 'type': 'string'},
        {'flag': 'public_dataset', 'type': 'boolean'},
        {'flag': 'black_box', 'type': 'boolean'},
        {'flag': 'white_box', 'type': 'boolean'},
        {'flag': 'model_price', 'type': 'float'},
        {'flag': 'dataset_price', 'type': 'float'},
        {'flag': 'cpp', 'type': 'float'},
        {'flag': 'progress_bar', 'type': 'boolean'},
        {'flag': 'resources_log', 'type': 'string'},
        {'flag': 'from_file', 'type': 'string'},
        {'flag': 'source_tag', 'type': 'string'},
        {'flag': 'dataset_tag', 'type': 'string'},
        {'flag': 'prediction_tag', 'type': 'string'},
        {'flag': 'evaluation_tag', 'type': 'string'},
        {'flag': 'ensemble_tag', 'type': 'string'},
        {'flag': 'all_tag', 'type': 'string'},
        {'flag': 'locale', 'type': 'string'},
        {'flag': 'combine_votes', 'type': 'string'},
        {'flag': 'plurality', 'type': 'string'},
        {'flag': 'verbosity', 'type': 'int'},
        {'flag': 'fields_map', 'type': 'string'},
        {'flag': 'clear_logs', 'type': 'boolean'},
        {'flag': 'cross_validation_rate', 'type': 'float'},
        {'flag': 'number_of_evaluations', 'type': 'int'},
        {'flag': 'store', 'type': 'boolean'},
        {'flag': 'test_split', 'type': 'float'},
        {'flag': 'ensemble', 'type': 'string'},
        {'flag': 'ensemble_file', 'type': 'string'},
        {'flag': 'tlp', 'type': 'int'},
        {'flag': 'prediction_info', 'type': 'string'},
        {'flag': 'max_parallel_evaluations', 'type': 'int'},
        {'flag': 'test_separator', 'type': 'string'},
        {'flag': 'multi_label', 'type': 'boolean'},
        {'flag': 'labels', 'type': 'string'},
        {'flag': 'label_separator', 'type': 'string'},
        {'flag': 'training_separator', 'type': 'string'},
        {'flag': 'prediction_header', 'type': 'boolean'},
        {'flag': 'prediction_fields', 'type': 'string'},
        {'flag': 'seed', 'type': 'string'},
        {'flag': 'max_parallel_ensembles', 'type': 'int'},
        {'flag': 'ensembles', 'type': 'string'},
        {'flag': 'threshold', 'type': 'int'},
        {'flag': 'threshold_class', 'type': 'string'},
        {'flag': 'max_categories', 'type': 'int'},
        {'flag': 'test_field_attributes', 'type': 'string'},
        {'flag': 'test_types', 'type': 'string'},
        {'flag': 'test_source', 'type': 'string'},
        {'flag': 'test_dataset', 'type': 'string'},
        {'flag': 'no_batch', 'type': 'boolean'},
        {'flag': 'dataset_attributes', 'type': 'string'},
        {'flag': 'output', 'type': 'string'},
        {'flag': 'new_fields', 'type': 'string'},
        {'flag': 'model_attributes', 'type': 'string'},
        {'flag': 'node_threshold', 'type': 'int'},
        {'flag': 'multi_label_fields', 'type': 'string'},
        {'flag': 'ensemble_attributes', 'type': 'string'},
        {'flag': 'source_attributes', 'type': 'string'},
        {'flag': 'evaluation_attributes', 'type': 'string'},
        {'flag': 'batch_prediction_attributes', 'type': 'string'},
        {'flag': 'batch_prediction_tag', 'type': 'string'},
        {'flag': 'balance', 'type': 'boolean'},
        {'flag': 'weight_field', 'type': 'string'},
        {'flag': 'objective_weights', 'type': 'string'},
        {'flag': 'label_aggregates', 'type': 'string'},
        {'flag': 'missing_strategy', 'type': 'string'},
        {'flag': 'other_than', 'type': 'string'},
        {'flag': 'newer_than', 'type': 'string'},
        {'flag': 'multi_dataset', 'type': 'boolean'},
        {'flag': 'multi_dataset_attributes', 'type': 'string'},
        {'flag': 'shared', 'type': 'boolean'},
        {'flag': 'reports', 'type': 'string'},
        {'flag': 'upload', 'type': 'boolean'},
        {'flag': 'test_dataset', 'type': 'string'},
        {'flag': 'dataset_off', 'type': 'boolean'},
        {'flag': 'args_separator', 'type': 'string'},
        {'flag': 'cluster_tag', 'type': 'string'},
        {'flag': 'centroid_tag', 'type': 'string'},
        {'flag': 'batch_centroid_tag', 'type': 'string'},
        {'flag': 'to_csv', 'type': 'string'},
        {'flag': 'resource_types', 'type': 'string'},
        {'flag': 'dry_run', 'type': 'boolean'},
        {'flag': 'anomaly_tag', 'type': 'string'},
        {'flag': 'anomaly_score_tag', 'type': 'string'},
        {'flag': 'project_tag', 'type': 'string'},
        {'flag': 'fast', 'type': 'boolean'},
        {'flag': 'project', 'type': 'string'},
        {'flag': 'project_id', 'type': 'string'},
        {'flag': 'no_csv', 'type': 'boolean'},
        {'flag': 'to_dataset', 'type': 'boolean'},
        {'flag': 'median', 'type': 'boolean'},
        {'flag': 'random_candidates', 'type': 'int'},
        {'flag': 'status', 'type': 'string'}],
    'BigMLer analyze': [
        {'flag': 'k-fold', 'type': 'int'},
        {'flag': 'cv', 'type': 'boolean'},
        {'flag': 'features', 'type': 'boolean'},
        {'flag': 'maximize', 'type': 'string'},
        {'flag': 'optimize', 'type': 'string'},
        {'flag': 'nodes', 'type': 'boolean'},
        {'flag': 'max_nodes', 'type': 'int'},
        {'flag': 'min_nodes', 'type': 'int'},
        {'flag': 'nodes_step', 'type': 'int'},
        {'flag': 'random_fields', 'type': 'boolean'},
        {'flag': 'exclude_features', 'type': 'string'},
        {'flag': 'optimize_category', 'type': 'string'}],
    'BigMLer cluster': [
        {'flag': 'cluster_fields', 'type': 'string'},
        {'flag': 'cluster', 'type': 'string'},
        {'flag': 'cluster_file', 'type': 'string'},
        {'flag': 'clusters', 'type': 'string'},
        {'flag': 'k', 'type': 'int'},
        {'flag': 'no_cluster', 'type': 'boolean'},
        {'flag': 'cluster_attributes', 'type': 'string'},
        {'flag': 'centroid_attributes', 'type': 'string'},
        {'flag': 'batch_centroid_attributes', 'type': 'string'},
        {'flag': 'cluster_datasets', 'type': 'string'},
        {'flag': 'cluster_models', 'type': 'string'},
        {'flag': 'summary_fields', 'type': 'string'}],
    'BigMLer anomaly': [
        {'flag': 'anomaly_fields', 'type': 'string'},
        {'flag': 'anomaly', 'type': 'string'},
        {'flag': 'anomaly_file', 'type': 'string'},
        {'flag': 'anomalies', 'type': 'string'},
        {'flag': 'no_anomaly', 'type': 'boolean'},
        {'flag': 'anomaly_attributes', 'type': 'string'},
        {'flag': 'anomaly_score_attributes', 'type': 'string'},
        {'flag': 'batch_anomaly_score_attributes', 'type': 'string'},
        {'flag': 'score', 'type': 'boolean'},
        {'flag': 'anomalies-dataset', 'type': 'string'},
        {'flag': 'top_n', 'type': 'int'},
        {'flag': 'forest_size', 'type': 'int'}],
    'BigMLer sample': [
        {'flag': 'anomaly_fields', 'type': 'string'},
        {'flag': 'sample', 'type': 'string'},
        {'flag': 'sample_file', 'type': 'string'},
        {'flag': 'samples', 'type': 'string'},
        {'flag': 'no_sample', 'type': 'boolean'},
        {'flag': 'sample_attributes', 'type': 'string'},
        {'flag': 'sample_header', 'type': 'boolean'},
        {'flag': 'fields_filter', 'type': 'string'},
        {'flag': 'row_index', 'type': 'int'},
        {'flag': 'mode', 'type': 'string'},
        {'flag': 'occurrence', 'type': 'int'},
        {'flag': 'precision', 'type': 'int'},
        {'flag': 'rows', 'type': 'int'},
        {'flag': 'row_offset', 'type': 'int'},
        {'flag': 'row_order_by', 'type': 'string'},
        {'flag': 'row_fields', 'type': 'string'},
        {'flag': 'stat_fields', 'type': 'string'},
        {'flag': 'stat_field', 'type': 'string'},
        {'flag': 'unique', 'type': 'boolean'}],
    'BigMLer report': [
        {'flag': 'from_dir', 'type': 'string'},
        {'flag': 'port', 'type': 'int'},
        {'flag': 'no_server', 'type': 'boolean'}],
    'BigMLer reify': [
        {'flag': 'language', 'type': 'string'},
        {'flag': 'add_fields', 'type': 'boolean'}],
    'BigMLer project': [
        {'flag': 'project_attributes', 'type': 'string'}],
    'BigMLer association': [
        {'flag': 'association_fields', 'type': 'string'},
        {'flag': 'association', 'type': 'string'},
        {'flag': 'association_file', 'type': 'string'},
        {'flag': 'associations', 'type': 'string'},
        {'flag': 'association_k', 'type': 'int'},
        {'flag': 'no_association', 'type': 'boolean'},
        {'flag': 'association_attributes', 'type': 'string'}]}


def get_user_defaults(defaults_file=DEFAULTS_FILE):
    """Looks for a defaults file and returns its contents

    """
    if defaults_file is None:
        defaults_file = DEFAULTS_FILE
    try:
        open(defaults_file).close()
        config = ConfigParser.ConfigParser()
        config.read(defaults_file)
        defaults = parse_user_defaults(config)
    except IOError:
        defaults = {}
        for section in FLAGS:
            defaults[section] = {}
    return defaults


def parse_user_defaults(config):
    """Reads default values from file

    """
    config_get = {'boolean': config.getboolean,
                  'float': config.getfloat,
                  'int': config.getint,
                  'string': config.get}
    defaults = {}
    for section in FLAGS:
        defaults[section] = {}
        for argument in FLAGS[section]:
            try:
                value = config_get[argument['type']](section,
                                                     argument['flag'])
                defaults[section].update({argument['flag']: value})
            except ConfigParser.Error:
                pass
    return defaults



# -*- coding: utf-8 -*-
#
# Copyright 2015 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""RESTCall class to store the REST API call definitions

"""

from __future__ import absolute_import

import pprint

from bigml.constants import RENAMED_RESOURCES
from bigml.resourcehandler import get_resource_type

INDENT = 4 * " "


def sort_lists(dict_structure):
    """Sorts the lists in a dict_structure

    """
    if dict_structure is not None and isinstance(dict_structure, dict):
        for key, value in dict_structure.items():
            if value is not None:
                if isinstance(value, list):
                    dict_structure[key] = sorted(value)
                elif isinstance(value, dict):
                    sort_lists(dict_structure[key])
    return dict_structure


class RESTCall(object):
    """Object to store the REST call definition

    """
    def __init__(self, action, origins=None, args=None,
                 resource_id=None, resource_type=None, suffix=None):
        """Constructor for the REST call definition

            resource_id: ID for the generated resource
            action: ["create" | "update" | "get"]
            origins: list of resouce IDs for the origin resources
            args: dict for the rest of arguments

        """
        self.action = action
        self.origins = [resource_id] if origins is None and \
                       action == "update" else origins
        self.args = args or {}
        input_data = self.args.get("input_data")
        if input_data:
            del self.args["input_data"]
        self.input_data = input_data
        self.resource_id = resource_id
        self.resource_type = resource_type
        self.suffix = suffix


    def reify(self, language=None, alias=None):
        """REST call command line

            language: computing language to write the output in
            alias: list of aliases for the related resources (e.g
                    {"source/55c4de8d1fa89c2dc70012d0": "source1"}
                    will cause the references to
                    "source/55c4de8d1fa89c2dc70012d0" to be renamed
                    as "source1")
        """
        out = ""
        if not language:
            out += "\n\n"
            out += "resource ID: %s\n" % self.resource_id
            out += "action: %s\n" % self.action
            out += "origins: %s\n" % self.origins
            out += "args: %s\n" % self.args
            out += "input data: %s\n" % self.input_data
            return out
        else:
            try:
                reify_handler = getattr(self, "reify_%s" % language)
            except AttributeError:
                reify_handler = self.reify
            return reify_handler(
                alias=alias)

    def reify_python(self, alias=None):
        """REST call command line in python. See ``reify`` method.

        """

        def resource_alias(resource_id):
            """Returns the alias if found

            """
            if isinstance(resource_id, basestring):
                return alias.get(resource_id, '"%s"' % resource_id)
            elif isinstance(resource_id, list):
                alias_names = []
                for resource_id_id in resource_id:
                    alias_names.append(
                        alias.get(resource_id_id, '"%s"' % resource_id_id))
                return repr(alias_names)

        resource_type = get_resource_type(self.resource_id)
        resource_name = resource_alias(self.resource_id)
        resource_method_suffix = RENAMED_RESOURCES.get(
            resource_type, resource_type)
        origin_names = [resource_alias(resource_id) for resource_id
                        in self.origins]

        arguments = ", ".join(origin_names)
        if self.suffix:
            arguments = "%s%s" % (arguments, self.suffix)
        if self.input_data:
            arguments = "%s, \\\n%s%s" % ( \
                arguments, INDENT,
                pprint.pformat(self.input_data).replace("\n", "\n%s" % INDENT))
        if self.args:
            sort_lists(self.args)
            arguments = "%s, \\\n%s%s" % (arguments, \
                INDENT, \
                pprint.pformat(self.args).replace( \
                    "\n", "\n%s" % INDENT))
        out = "%s = api.%s_%s(%s)\napi.ok(%s)\n\n" % (
            resource_name,
            self.action,
            resource_method_suffix,
            arguments,
            resource_name)
        return out



# -*- coding: utf-8 -*-
#
# Copyright 2012-2015 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Prediction auxiliary functions

"""
from __future__ import absolute_import

import sys
import ast
import gc

import bigml.api

import bigmler.utils as u
import bigmler.checkpoint as c



from bigml.model import Model
from bigml.multimodel import MultiModel, read_votes
from bigml.ensemble import Ensemble
from bigml.util import localize, console_log, get_predictions_file_name
from bigml.io import UnicodeWriter
from bigml.multivote import (PLURALITY_CODE, THRESHOLD_CODE, MultiVote,
                             ws_confidence)

from bigmler.tst_reader import TstReader as TestReader
from bigmler.resources import (FIELDS_QS, ALL_FIELDS_QS, BRIEF_FORMAT,
                               NORMAL_FORMAT, FULL_FORMAT)
from bigmler.resources import create_batch_prediction

MAX_MODELS = 10
AGGREGATION = -1
COMBINATION = -2
COMBINATION_LABEL = 'combined'
OTHER = "***** other *****"


def use_prediction_headers(prediction_headers, output, test_reader,
                           fields, args, objective_field):
    """Uses header information from the test file in the prediction output

       If --prediction-header is set, adds a headers row to the prediction
       file.
       If --prediction-fields is used, retrieves the fields to exclude
       from the test input in the --prediction-info full format, that includes
       them all by default.

    """
    exclude = []
    try:
        if not objective_field in fields.fields:
            objective_field = fields.field_id(objective_field)
        objective_name = fields.field_name(objective_field)
    except ValueError, exc:
        sys.exit(exc)
    headers = [objective_name]

    if args.prediction_info in [NORMAL_FORMAT, FULL_FORMAT]:
        headers.append("confidence")
    if (args.prediction_info == FULL_FORMAT or
            args.prediction_fields is not None):
        # Try to retrieve headers from the test file
        if test_reader.has_headers():
            input_headers = test_reader.raw_headers[:]
            if objective_name in input_headers:
                exclude.append(input_headers.index(objective_name))
        else:
            # if no headers are found in the test file we assume it has the
            # same model input_field structure removing the objective field
            input_headers = [fields[field]['name'] for field in
                             fields.fields_columns if not
                             field == objective_field]

        if args.prediction_fields is not None:
            prediction_fields = [field.strip() for field in
                                 args.prediction_fields.split(',')]
            # Filter input_headers adding only those chosen by the user
            number_of_headers = len(input_headers)
            for index in range(0, number_of_headers):
                if not input_headers[index] in prediction_fields:
                    exclude.append(index)
        exclude = sorted(list(set(exclude)), reverse=True)
        for index in exclude:
            del input_headers[index]
        input_headers.extend(headers)
        headers = input_headers
    if prediction_headers:
        output.writerow(headers)
    return exclude


def write_prediction(prediction, output=sys.stdout,
                     prediction_info=NORMAL_FORMAT, input_data=None,
                     exclude=None):
    """Writes the final combined prediction to the required output

       The format of the output depends on the `prediction_info` value.
       There's a brief format, that writes only the predicted value,
       a normal format (default) that writes the prediction followed by its
       confidence, and a full data format that writes first the input data
       used to predict followed by the prediction.

    """
    confidence = False
    # handles (prediction, confidence) input
    if isinstance(prediction, tuple):
        prediction, confidence = prediction
    # handles [prediction] or [prediction, confidence, ...] input
    if isinstance(prediction, list):
        prediction, confidence = ((prediction[0], None) if len(prediction) == 1
                                  else prediction)

    row = []
    # input data is added if prediction format is BRIEF (no confidence) or FULL
    if prediction_info != NORMAL_FORMAT:
        if input_data is None:
            input_data = []
        row = input_data
        if exclude:
            for index in exclude:
                del row[index]
    row.append(prediction)
    if prediction_info in [NORMAL_FORMAT, FULL_FORMAT]:
        row.append(confidence)
    try:
        output.writerow(row)
    except AttributeError:
        try:
            output.write(row)
        except AttributeError:
            raise AttributeError("You should provide a writeable object")


def prediction_to_row(prediction, prediction_info=NORMAL_FORMAT):
    """Returns a csv row to store main prediction info in csv files.

    """
    prediction = prediction['object']
    prediction_class = prediction['objective_fields'][0]
    tree = prediction.get('prediction_path', prediction)
    row = [prediction['prediction'][prediction_class]]
    if not prediction_info == BRIEF_FORMAT:
        row.append(prediction.get('confidence', tree.get('confidence', 0)))
        distribution = None
        if 'objective_summary' in tree:
            summary = tree['objective_summary']
            if 'bins' in summary:
                distribution = summary['bins']
            elif 'counts' in summary:
                distribution = summary['counts']
            elif 'categories' in summary:
                distribution = summary['categories']
        if distribution:
            row.extend([repr(distribution), sum([x[1] for x in distribution])])
    return row


def combine_votes(votes_files, to_prediction, to_file, method=0,
                  prediction_info=NORMAL_FORMAT, input_data_list=None,
                  exclude=None):
    """Combines the votes found in the votes' files and stores predictions.

       votes_files: should contain the list of file names
       to_prediction: is the Model method that casts prediction to numeric
                      type if needed
       to_file: is the name of the final output file.
    """
    votes = read_votes(votes_files, to_prediction)

    u.check_dir(to_file)
    with UnicodeWriter(to_file) as output:
        number_of_tests = len(votes)
        if input_data_list is None or len(input_data_list) != number_of_tests:
            input_data_list = None
        for index in range(0, number_of_tests):
            multivote = votes[index]
            input_data = (None if input_data_list is None
                          else input_data_list[index])
            write_prediction(multivote.combine(method, True), output,
                             prediction_info, input_data, exclude)


def remote_predict_models(models, test_reader, prediction_file, api, args,
                          resume=False, output_path=None,
                          session_file=None, log=None, exclude=None):
    """Retrieve predictions remotely, combine them and save predictions to file

    """
    predictions_files = []
    prediction_args = {
        "tags": args.tag
    }
    test_set_header = test_reader.has_headers()
    if output_path is None:
        output_path = u.check_dir(prediction_file)
    message_logged = False

    raw_input_data_list = []
    for input_data in test_reader:
        raw_input_data_list.append(input_data)
    single_model = len(models) == 1
    if single_model:
        prediction_file = UnicodeWriter(prediction_file).open_writer()
    for model in models:
        model = bigml.api.get_model_id(model)
        predictions_file = get_predictions_file_name(model,
                                                     output_path)
        predictions_files.append(predictions_file)
        if (not resume or
                not c.checkpoint(c.are_predictions_created, predictions_file,
                                 test_reader.number_of_tests(),
                                 debug=args.debug)[0]):
            if not message_logged:
                message = u.dated("Creating remote predictions.\n")
                u.log_message(message, log_file=session_file,
                              console=args.verbosity)
            message_logged = True
            with UnicodeWriter(predictions_file) as predictions_file:
                for input_data in raw_input_data_list:
                    input_data_dict = test_reader.dict(input_data)
                    prediction = api.create_prediction(model, input_data_dict,
                                                       by_name=test_set_header,
                                                       wait_time=0,
                                                       args=prediction_args)
                    u.check_resource_error(prediction,
                                           "Failed to create prediction: ")
                    u.log_message("%s\n" % prediction['resource'],
                                  log_file=log)
                    prediction_row = prediction_to_row(prediction)
                    predictions_file.writerow(prediction_row)
                    if single_model:
                        write_prediction(prediction_row[0:2], prediction_file,
                                         args.prediction_info,
                                         input_data, exclude)
    if single_model:
        prediction_file.close_writer()
    else:
        combine_votes(predictions_files,
                      Model(models[0]).to_prediction,
                      prediction_file, args.method,
                      args.prediction_info, raw_input_data_list, exclude)


def remote_predict_ensemble(ensemble_id, test_reader, prediction_file, api,
                            args, resume=False, output_path=None,
                            session_file=None, log=None, exclude=None):
    """Retrieve predictions remotely and save predictions to file

    """
    prediction_args = {
        "tags": args.tag,
        "combiner": args.method
    }
    test_set_header = test_reader.has_headers()
    if output_path is None:
        output_path = u.check_dir(prediction_file)

    if (not resume or not c.checkpoint(
            c.are_predictions_created, prediction_file,
            test_reader.number_of_tests(), debug=args.debug)[0]):
        message = u.dated("Creating remote predictions.")
        u.log_message(message, log_file=session_file,
                      console=args.verbosity)

        with UnicodeWriter(prediction_file) as predictions_file:
            for input_data in test_reader:
                input_data_dict = test_reader.dict(input_data)
                prediction = api.create_prediction(ensemble_id,
                                                   input_data_dict,
                                                   by_name=test_set_header,
                                                   wait_time=0,
                                                   args=prediction_args)
                prediction = u.check_resource(prediction,
                                              api.get_prediction)
                u.check_resource_error(prediction,
                                       "Failed to create prediction: ")
                u.log_message("%s\n" % prediction['resource'], log_file=log)
                prediction_row = prediction_to_row(prediction,
                                                   args.prediction_info)
                write_prediction(prediction_row, predictions_file,
                                 args.prediction_info, input_data, exclude)


def local_predict(models, test_reader, output, args, options=None,
                  exclude=None):
    """Get local predictions and combine them to get a final prediction

    """
    single_model = len(models) == 1
    test_set_header = test_reader.has_headers()
    kwargs = {"by_name": test_set_header, "with_confidence": True,
              "missing_strategy": args.missing_strategy}
    if single_model:
        local_model = Model(models[0])
    else:
        local_model = Ensemble(models, max_models=args.max_batch_models)
        kwargs.update({"method": args.method, "options": options,
                       "median": args.median})
    for input_data in test_reader:
        input_data_dict = dict(zip(test_reader.raw_headers, input_data))
        prediction = local_model.predict(
            input_data_dict, **kwargs)
        if single_model and args.median and local_model.tree.regression:
            # only single models' predictions can be based on the median value
            # predict
            prediction[0] = prediction[-1]
        write_prediction(prediction[0: 2],
                         output,
                         args.prediction_info, input_data, exclude)


def retrieve_models_split(models_split, api, query_string=FIELDS_QS,
                          labels=None, multi_label_data=None, ordered=True,
                          models_order=None):
    """Returns a list of full model structures ready to be fed to the
       MultiModel object to produce predictions. Models are also stored
       locally in the output directory when the --store flag is used.

    """
    complete_models = []
    if models_order is None:
        models_order = []
    for index in range(len(models_split)):
        model = models_split[index]
        if (isinstance(model, basestring) or
                bigml.api.get_status(model)['code'] != bigml.api.FINISHED):
            try:
                model = u.check_resource(model, api.get_model,
                                         query_string)
            except ValueError, exception:
                sys.exit("Failed to get model: %s. %s" % (model,
                                                          str(exception)))

        # When user selects the labels in multi-label predictions, we must
        # filter the models that will be used to predict
        if labels and multi_label_data:
            objective_column = str(multi_label_data['objective_column'])
            labels_info = multi_label_data[
                'generated_fields'][objective_column]
            labels_columns = [label_info[1] for label_info in labels_info
                              if label_info[0] in labels]
            model_objective_id = model['object']['objective_fields'][0]
            model_fields = model['object']['model']['fields']
            model_objective = model_fields[model_objective_id]
            model_column = model_objective['column_number']
            if model_column in labels_columns:
                # When the list of models comes from a --model-tag
                # selection, the models are not retrieved in the same
                # order they were created. We must keep track of the
                # label they are associated with to label their
                # predictions properly
                if not ordered:
                    models_order.append(model_column)
                complete_models.append(model)
        else:
            complete_models.append(model)
    return complete_models, models_order


def aggregate_multivote(multivote, options, labels, models_per_label, ordered,
                        models_order, label_separator=None):
    """Aggregate the model's predictions for multi-label fields in a
       concatenated format into a final prediction

    """

    if label_separator is None:
        label_separator = ","
    predictions = multivote.predictions

    if ordered and models_per_label == 1:
        # as multi-labeled models are created from end to start votes
        # must be reversed to match
        predictions.reverse()
    else:
        predictions = [prediction for (_, prediction)
                       in sorted(zip(models_order, predictions),
                                 key=lambda x: x[0])]

    if (labels is None or
            len(labels) * models_per_label != len(predictions)):
        sys.exit("Failed to make a multi-label prediction. No"
                 " valid label info is found.")
    prediction_list = []
    confidence_list = []
    # In the following case, we must vote each label using the models
    # in the ensemble and the chosen method

    if models_per_label > 1:
        label_predictions = [predictions[i: i + models_per_label] for
                             i in range(0, len(predictions),
                                        models_per_label)]
        predictions = []
        for label_prediction in label_predictions:
            label_multivote = MultiVote(label_prediction)
            prediction, confidence = label_multivote.combine(
                method=AGGREGATION, with_confidence=True, options=options)
            predictions.append({'prediction': prediction,
                                'confidence': confidence})
    for vote_index in range(0, len(predictions)):
        if ast.literal_eval(predictions[vote_index]['prediction']):
            prediction_list.append(labels[vote_index])
            confidence = str(predictions[vote_index]['confidence'])
            confidence_list.append(confidence)
    prediction = [label_separator.join(prediction_list),
                  label_separator.join(confidence_list)]
    return prediction


def combine_multivote(multivote, other_label=OTHER):
    """Combine in a global distribution the distribution of predictions
       obtained with models when each one is built on a subset of training
       data that has a subset of categories.

    """
    predictions = multivote.predictions
    global_distribution = []
    for prediction in predictions:
        prediction_category = None
        prediction_instances = 0
        for category, instances in prediction['distribution']:
            if category != other_label:
                if instances > prediction_instances:
                    prediction_category = category
                    prediction_instances = instances
        if prediction_category is not None:
            prediction_confidence = ws_confidence(
                prediction_category, prediction['distribution'])
            global_distribution.append([prediction_category,
                                        prediction_confidence])
    if global_distribution:
        prediction = sorted(global_distribution, key=lambda x: x[1],
                            reverse=True)[0]
    else:
        prediction = [None, None]
    return prediction


def local_batch_predict(models, test_reader, prediction_file, api, args,
                        resume=False, output_path=None, output=None,
                        method=PLURALITY_CODE, options=None,
                        session_file=None, labels=None, ordered=True,
                        exclude=None, models_per_label=1, other_label=OTHER,
                        multi_label_data=None):

    """Get local predictions form partial Multimodel, combine and save to file

    """

    def draw_progress_bar(current, total):
        """Draws a text based progress report.

        """
        pct = 100 - ((total - current) * 100) / (total)
        console_log("Predicted on %s out of %s models [%s%%]" % (
            localize(current), localize(total), pct), reset=True)

    max_models = args.max_batch_models
    if labels is None:
        labels = []
    test_set_header = test_reader.has_headers()
    if output_path is None:
        output_path = u.check_dir(prediction_file)
    if output is None:
        try:
            output = open(prediction_file, 'w', 0)
        except IOError:
            raise IOError("Failed to write in %s" % prediction_file)
    models_total = len(models)
    models_splits = [models[index:(index + max_models)] for index
                     in range(0, models_total, max_models)]
    # Input data is stored as a list and predictions are made for all rows
    # with each model
    raw_input_data_list = []
    for input_data in test_reader:
        raw_input_data_list.append(input_data)
    total_votes = []
    models_order = []
    models_count = 0
    single_model = models_total == 1
    query_string = FIELDS_QS if single_model else ALL_FIELDS_QS
    # processing the models in slots
    for models_split in models_splits:
        if resume:
            for model in models_split:
                pred_file = get_predictions_file_name(model,
                                                      output_path)
                c.checkpoint(c.are_predictions_created,
                             pred_file,
                             test_reader.number_of_tests(), debug=args.debug)
        # retrieving the full models allowed by --max-batch-models to be used
        # in a multimodel slot
        complete_models, models_order = retrieve_models_split(
            models_split, api, query_string=query_string, labels=labels,
            multi_label_data=multi_label_data, ordered=ordered,
            models_order=models_order)

        # predicting with the multimodel slot
        if complete_models:
            local_model = MultiModel(complete_models, api=api)
            # added to ensure garbage collection at each step of the loop
            gc.collect()
            try:
                votes = local_model.batch_predict(
                    raw_input_data_list, output_path, by_name=test_set_header,
                    reuse=True, missing_strategy=args.missing_strategy,
                    headers=test_reader.raw_headers, to_file=(not args.fast),
                    use_median=args.median)
            except ImportError:
                sys.exit("Failed to find the numpy and scipy libraries needed"
                         " to use proportional missing strategy for"
                         " regressions. Please, install them manually")

            # extending the votes for each input data with the new model-slot
            # predictions
            if not args.fast:
                votes = local_model.batch_votes(output_path)
            models_count += max_models
            if models_count > models_total:
                models_count = models_total
            if args.verbosity:
                draw_progress_bar(models_count, models_total)

            if total_votes:
                for index in range(0, len(votes)):
                    predictions = total_votes[index]
                    predictions.extend(votes[index].predictions)
            else:
                total_votes = votes

    if not single_model:
        message = u.dated("Combining predictions.\n")
        u.log_message(message, log_file=session_file, console=args.verbosity)

    # combining the votes to issue the final prediction for each input data
    for index in range(0, len(total_votes)):
        multivote = total_votes[index]
        input_data = raw_input_data_list[index]

        if single_model:
            # single model predictions need no combination
            prediction = [multivote.predictions[0]['prediction'],
                          multivote.predictions[0]['confidence']]
        elif method == AGGREGATION:
            # multi-labeled fields: predictions are concatenated
            prediction = aggregate_multivote(
                multivote, options, labels, models_per_label, ordered,
                models_order, label_separator=args.label_separator)
        elif method == COMBINATION:
            # used in --max-categories flag: each model slot contains a
            # subset of categories and the predictions for all of them
            # are combined in a global distribution to obtain the final
            # prediction
            prediction = combine_multivote(multivote, other_label=other_label)
        else:
            prediction = multivote.combine(method=method, with_confidence=True,
                                           options=options)

        write_prediction(prediction, output, args.prediction_info, input_data,
                         exclude)


def predict(models, fields, args, api=None, log=None,
            resume=False, session_file=None,
            labels=None, models_per_label=1, other_label=OTHER,
            multi_label_data=None):
    """Computes a prediction for each entry in the `test_set`.

       Predictions computed locally using MultiModels on subgroups of models.
       Chosing a max_batch_models value not bigger than the number_of_models
       flag will lead to the last case, where memory usage is bounded and each
       model predictions are saved for further use.
    """
    test_set = args.test_set
    test_set_header = args.test_header
    objective_field = args.objective_field
    output = args.predictions
    test_reader = TestReader(test_set, test_set_header, fields,
                             objective_field,
                             test_separator=args.test_separator)

    prediction_file = output
    output_path = u.check_dir(output)
    with UnicodeWriter(output) as output:
        # columns to exclude if input_data is added to the prediction field
        exclude = use_prediction_headers(
            args.prediction_header, output, test_reader, fields, args,
            objective_field)

        # Remote predictions: predictions are computed in bigml.com and stored
        # in a file named after the model in the following syntax:
        #     model_[id of the model]__predictions.csv
        # For instance,
        #     model_50c0de043b563519830001c2_predictions.csv
        # Predictions are computed individually only if no_batch flag is set
        if (args.remote and args.no_batch and not args.multi_label
                and args.method != THRESHOLD_CODE):
            if args.ensemble is not None:
                remote_predict_ensemble(args.ensemble, test_reader,
                                        prediction_file, api, args, resume,
                                        output_path, session_file, log,
                                        exclude)
            else:
                remote_predict_models(models, test_reader, prediction_file,
                                      api, args, resume, output_path,
                                      session_file, log, exclude)
            return
        # Local predictions: Predictions are computed locally using models'
        # rules with MultiModel's predict method
        message = u.dated("Creating local predictions.\n")
        u.log_message(message, log_file=session_file, console=args.verbosity)
        options = {}
        if args.method == THRESHOLD_CODE:
            options.update(threshold=args.threshold)
            if args.threshold_class is None:
                local_model = Model(models[0])
                args.threshold_class = local_model.tree.distribution[0][0]
            options.update(category=args.threshold_class)
        # For a model we build a Model and for a small number of models,
        # we build a MultiModel using all of
        # the given models and issue a combined prediction
        if (len(models) <= args.max_batch_models and args.fast and
                not args.multi_label and args.max_categories == 0
                and args.method != COMBINATION):
            local_predict(models, test_reader, output, args, options, exclude)
        # For large numbers of models, we split the list of models in chunks
        # and build a MultiModel for each chunk, issue and store predictions
        # for each model and combine all of them eventually.
        else:
            # Local predictions: predictions are computed locally using
            # models' rules with MultiModel's predict method and combined using
            # aggregation if the objective field is a multi-labelled field
            # or one of the available combination methods: plurality,
            # confidence weighted and probability weighted
            if args.multi_label:
                method = AGGREGATION
            elif args.max_categories > 0:
                method = COMBINATION
            else:
                method = args.method

            # For multi-labelled models, the --models flag keeps the order
            # of the labels and the models but the --model-tag flag
            # retrieves the models with no order, so the correspondence with
            # each label must be restored.
            ordered = True
            if args.multi_label and (args.model_tag is not None
                                     or models_per_label > 1):
                ordered = False
            local_batch_predict(models, test_reader, prediction_file, api,
                                args, resume=resume,
                                output_path=output_path,
                                output=output, method=method, options=options,
                                session_file=session_file, labels=labels,
                                ordered=ordered, exclude=exclude,
                                models_per_label=models_per_label,
                                other_label=other_label,
                                multi_label_data=multi_label_data)
    test_reader.close()


def remote_predict(model, test_dataset, batch_prediction_args, args,
                   api, resume, prediction_file=None, session_file=None,
                   path=None, log=None):
    """Computes a prediction for each entry in the `test_set`.

       Predictions are computed remotely using the batch predictions call.
    """

    if args.ensemble is not None:
        model_or_ensemble = args.ensemble
    else:
        model_or_ensemble = bigml.api.get_model_id(model)
    # if resuming, try to extract dataset form log files
    if resume:
        message = u.dated("Batch prediction not found. Resuming.\n")
        resume, batch_prediction = c.checkpoint(
            c.is_batch_prediction_created, path, debug=args.debug,
            message=message, log_file=session_file, console=args.verbosity)
    if not resume:
        batch_prediction = create_batch_prediction(
            model_or_ensemble, test_dataset, batch_prediction_args,
            args, api, session_file=session_file, path=path, log=log)
    if not args.no_csv:
        file_name = api.download_batch_prediction(batch_prediction,
                                                  prediction_file)
        if file_name is None:
            sys.exit("Failed downloading CSV.")
    if args.to_dataset:
        batch_prediction = bigml.api.check_resource(batch_prediction, api=api)
        new_dataset = bigml.api.get_dataset_id(
            batch_prediction['object']['output_dataset_resource'])
        if new_dataset is not None:
            message = u.dated("Batch prediction dataset created: %s\n"
                              % u.get_url(new_dataset))
            u.log_message(message, log_file=session_file,
                          console=args.verbosity)
            u.log_created_resources("batch_prediction_dataset",
                                    path, new_dataset, mode='a')
